{
	"namespace": "efte",
	"type": "record",
	"name": "alert",
	"doc": "avro alert schema for EFTE-generated alerts ",
	"version": "0.1",
	"fields": [
        {"name": "schemavsn", "type": "string", "doc": "schema version used"},
        {"name": "publisher", "type": "string", "doc": "origin of alert packet"},
		{"name": "objectId", "type": "string", "doc": "object identifier or name"},
		{"name": "candidate", "type": "candidate"},
		{"name": "xmatch", 
            "type": [
                "null" ,
                {
                "type": "array",
                "items": "xmatch"
                }
                ],
             "default": null}
	]
}